#include <iostream>
#include <stdlib.h>
#include "world.hh"

// overload + 
Rgb operator+(Rgb a, Rgb b)
{
  Rgb temp;
  temp.r = a.r + b.r;
  temp.g = a.g + b.g;
  temp.b = a.b + b.b;
  return temp;
}

World::World(short x, short y) : xsize(x), ysize(y)
{
  data = new guchar[(x + 2) * (y + 2) * 3];  // * 3 for RGB // +2 for border
}

World::~World()
{
  delete []data;
}

inline bool World::inbounds(short x, short y)
{
  if (x > xsize || x < 0 || y > ysize || y < 0) {
    cout << "Value out of bounds!" << endl;
    return FALSE;
  }
  return TRUE;
}

inline Rgb World::val(short x, short y)
{
  /*
  if (!(inbounds(x,y)))
    return 0;
  */
  Rgb rgb;
  rgb.r = data[3 * (x + y * (xsize + 2)) + 0];
  rgb.g = data[3 * (x + y * (xsize + 2)) + 1];
  rgb.b = data[3 * (x + y * (xsize + 2)) + 2];
  return rgb;
}

inline bool World::val(short x, short y, Rgb rgb)
{
  if (!(inbounds(x,y)))
    return FALSE;
  data[3 * (x + y * (xsize + 2)) + 0] = rgb.r;
  data[3 * (x + y * (xsize + 2)) + 1] = rgb.g;
  data[3 * (x + y * (xsize + 2)) + 2] = rgb.b;
  return TRUE;
}

inline bool World::val(short x, short y, char r, char g, char b)
{
  if (!(inbounds(x,y)))
    return FALSE;
  data[3 * (x + y * (xsize + 2)) + 0] = r;
  data[3 * (x + y * (xsize + 2)) + 1] = g;
  data[3 * (x + y * (xsize + 2)) + 2] = b;
  return TRUE;
}

bool World::row_cp(short y1, short y2)
{
  if (!(inbounds(0,y1)))
    return FALSE;
  if (!(inbounds(0,y2)))
    return FALSE;
  y1++; y2++;
  for (short x = 1; x <= xsize; x++) {
    data[3 * (x + y2 * (xsize + 2)) + 0] = data[3 * (x + y1 * (xsize + 2)) + 0];
    data[3 * (x + y2 * (xsize + 2)) + 1] = data[3 * (x + y1 * (xsize + 2)) + 1];
    data[3 * (x + y2 * (xsize + 2)) + 2] = data[3 * (x + y1 * (xsize + 2)) + 2];
  }
  return TRUE;
}

bool World::set_cell(short x, short y, char r, char g, char b)
{
  if (!(inbounds(x,y)))
    return FALSE;

  x++; y++;
  data[3 * (x + y * (xsize + 2)) + 0] = r;
  data[3 * (x + y * (xsize + 2)) + 1] = g;
  data[3 * (x + y * (xsize + 2)) + 2] = b;
  return TRUE;
}

bool World::fill(short x1, short y1, short x2, short y2, Rgb rgb)
{
  if (!(inbounds(x1,y1) && inbounds(x2,y2)))
    return FALSE;
  x1++; x2++; y1++; y2++;
  for (short y = y1; y < y2; y++)
    for (short x = x1; x < x2; x++) {
      data[3 * (x + y * (xsize + 2)) + 0] = rgb.r;
      data[3 * (x + y * (xsize + 2)) + 1] = rgb.g;
      data[3 * (x + y * (xsize + 2)) + 2] = rgb.b;
    }
  return TRUE;
}

bool World::fill(short x1, short y1, short x2, short y2, char r, char g, char b)
{
  if (!(inbounds(x1,y1) && inbounds(x2,y2)))
    return FALSE;
  x1++; x2++; y1++; y2++;
  for (short y = y1; y < y2; y++)
    for (short x = x1; x < x2; x++) {
      data[3 * (x + y * (xsize + 2)) + 0] = r;
      data[3 * (x + y * (xsize + 2)) + 1] = g;
      data[3 * (x + y * (xsize + 2)) + 2] = b;
    }
  return TRUE;
}

bool World::fill_rand(short x1, short y1, short x2, short y2, short chance)
{
  if (!(inbounds(x1,y1) && inbounds(x2,y2)))
    return FALSE;
  x1++; x2++; y1++; y2++;
  for (short y = y1; y < y2; y++)
    for (short x = x1; x < x2; x++) {
      if (rand() % chance == 0) {
	data[3 * (x + y * (xsize + 2)) + 0] = rand() % 256;
	data[3 * (x + y * (xsize + 2)) + 1] = rand() % 256;
	data[3 * (x + y * (xsize + 2)) + 2] = rand() % 256;
      }
    }
  return TRUE;
}

void World::cycle()
{
  Rgb rgb;
  for(int y = 1; y <= ysize; y++)
    for(int x = 1; x <= xsize; x++) {
      rgb = val(x - 1,y) + val(x + 1,y) + val(x,y + 1) + val(x,y - 1);
      val(x,y,rgb);
      /*
	cout << "World::cycle() says: " 
	<< (int)rgb.r << "," 
	<< (int)rgb.g << "," 
	<< (int)rgb.b << endl;
      */
    }
}

void World::run1disc1()
{
  char rules[2][2][2];
  rules[0][0][0] = 0;
  rules[0][0][1] = 255;
  rules[0][1][0] = 0;
  rules[0][1][1] = 0;
  rules[1][0][0] = 255;
  rules[1][0][1] = 0;
  rules[1][1][0] = 0;
  rules[1][1][1] = 0;
  
  bool a, b, c;
  for (short y = 2; y <= ysize; y++)
    for (short x = 1; x <= xsize; x++)
      for (short i = 0; i <= 2; i++) {
	a = b = c = 0;
	if (data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i])
	  a++;
	if (data[3 * ((x + 0) + (y - 1) * (xsize + 2)) + i])
	  b++;
	if (data[3 * ((x + 1) + (y - 1) * (xsize + 2)) + i])
	  c++;
	data[3 * (x + y * (xsize + 2)) + 0] = rules[a][b][c] - 128 * (rules[a][b][c] != 0);
	data[3 * (x + y * (xsize + 2)) + 1] = rules[a][b][c] -  64 * (rules[a][b][c] != 0);
	data[3 * (x + y * (xsize + 2)) + 2] = rules[a][b][c];
      }
}

void World::run1disc2()
{
  guchar rules[3][2][2][2];
  // red rules
  rules[1][0][0][0] = 0;
  rules[1][0][0][1] = 1;
  rules[1][0][1][0] = 1;
  rules[1][0][1][1] = 1;
  rules[1][1][0][0] = 0;
  rules[1][1][0][1] = 1;
  rules[1][1][1][0] = 1;
  rules[1][1][1][1] = 0;
  // green rules
  rules[0][0][0][0] = 0;
  rules[0][0][0][1] = 1;
  rules[0][0][1][0] = 1;
  rules[0][0][1][1] = 0;
  rules[0][1][0][0] = 1;
  rules[0][1][0][1] = 0;
  rules[0][1][1][0] = 0;
  rules[0][1][1][1] = 0;
  // blue rules
  rules[2][0][0][0] = 0;
  rules[2][0][0][1] = 1;
  rules[2][0][1][0] = 1;
  rules[2][0][1][1] = 1;
  rules[2][1][0][0] = 1;
  rules[2][1][0][1] = 0;
  rules[2][1][1][0] = 0;
  rules[2][1][1][1] = 0;
  
  bool a, b, c;
  int blah;
  guchar smack;
  short cnt;
  for (short y = 2; y <= ysize; y++)
    for (short x = 1; x <= xsize; x++)
      for (short i = 0; i <= 2; i++) {
	cnt = a = b = c = 0;
	if (data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i])
	  a++;
	if (data[3 * ((x + 0) + (y - 1) * (xsize + 2)) + i])
	  b++;
	if (data[3 * ((x + 1) + (y - 1) * (xsize + 2)) + i])
	  c++;
	cnt /= 3;
	//	if (cnt > 0) cout << "Cnt is " << cnt << endl;
	blah = 
	  data[3 * ((x + 1) + (y - 1) * (xsize + 2)) + i] +
	  data[3 * ((x + 0) + (y - 1) * (xsize + 2)) + i];
	//	blah >> 2;
	blah += data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i];
	//	blah >> 2;
	smack = static_cast<guchar>(blah);
	//	if (smack > 1) cout << "Smack me: " << static_cast<int>(smack) << endl;
	data[3 * (x + y * (xsize + 2)) + i] = smack * rules[i][a][b][c];
	
	//static_cast<guchar>(cnt); // * rules[i][a][b][c];
      }
}

void World::run1cont(World * parm)
{
  double newstate, neighbors, parm_neighbors;
  for (short y = 2; y <= ysize; y++) {
    //    cout << "On row " << y << endl;
    for (short x = 1; x <= xsize; x++)
      for (register short i = 0; i <= 2; i++) {
	neighbors = 
	  data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i]
	  //	  + data[3 * ((x + 0) + (y - 1) * (xsize + 2)) + i]
	  + data[3 * ((x + 1) + (y - 1) * (xsize + 2)) + i];
	neighbors /= 2.0;
	parm_neighbors = 
	  parm->data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i]
	  // + parm->data[3 * ((x + 0) + (y - 1) * (xsize + 2)) + i]
	  + parm->data[3 * ((x + 1) + (y - 1) * (xsize + 2)) + i];
	parm_neighbors /= 2.0;
	newstate = (255.0 * 4.9) * parm->data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i] * neighbors * (255.0 - neighbors);
	if (newstate > 255.0)
	  newstate = (255.0 * newstate) / (newstate + 255.0);
	data[3 * (x + y * (xsize + 2)) + i] = static_cast<guchar>(newstate);
	parm->data[3 * (x + y * (xsize + 2)) + i] = static_cast<guchar>((neighbors + parm_neighbors) / 2.0);
      }
  }
}

void World::run2cont(World * parm)
{
  double newstate, neighbors, parm_neighbors;
  for (short y = 1; y <= ysize; y++) {
    for (short x = 1; x <= xsize; x++)
      for (register short i = 0; i <= 2; i++) {
	
	neighbors = 
	  data[3 * ((x    ) + (y + 1) * (xsize + 2)) + i] +
	  data[3 * ((x    ) + (y - 1) * (xsize + 2)) + i] +
	  data[3 * ((x + 1) + (y    ) * (xsize + 2)) + i] +
	  data[3 * ((x - 1) + (y    ) * (xsize + 2)) + i];
	neighbors /= 4.0;
	
	parm_neighbors = 
	  parm->data[3 * ((x    ) + (y + 1) * (xsize + 2)) + i] +
	  parm->data[3 * ((x    ) + (y - 1) * (xsize + 2)) + i] +
	  parm->data[3 * ((x + 1) + (y    ) * (xsize + 2)) + i] +
	  parm->data[3 * ((x - 1) + (y    ) * (xsize + 2)) + i];
	parm_neighbors /= 4.0;

	newstate = (255.0 * 4.9) * parm->data[3 * ((x - 1) + (y - 1) * (xsize + 2)) + i] * neighbors * (255.0 - neighbors);
	//if (newstate > 255.0)
	//  newstate = (255.0 * newstate) / (newstate + 255.0);
	newstate = 255.0 * cos(newstate);
	data[3 * (x + y * (xsize + 2)) + i] = static_cast<guchar>(newstate);
	parm->data[3 * (x + y * (xsize + 2)) + i] = static_cast<guchar>((neighbors + parm_neighbors) / 2.0);
      }
  }
}  
